import os
import re
import tempfile
import sqlparse
from google.cloud import bigquery
from google.cloud import storage


def read_sql(sql_path,
             sql_params={}):
    """
    Reads a .sql file, removes comments, and (optional) substitutes a dict of parameters
    :param sql_path: path/to/script.sql
    :param sql_params: example {'param_1': 3, 'param_2': ['a','b','c']}
    :return: string
    """
    # read the sql, remove the comments:
    with open(sql_path, 'r') as f:
        sql = sqlparse.format(f.read(), strip_comments=True)

    # substitute parameters:
    for key in sql_params.keys():
        if isinstance(sql_params[key], list):
            if isinstance(sql_params[key][0], str):
                sql_params[key] = '\',\''.join(sql_params[key])
                sql_params[key] = '\'{}\''.format(sql_params[key])
            else:
                sql_params[key] = ','.join([str(item) for item in sql_params[key]])

        sql = re.sub(key, str(sql_params[key]), sql)

    return sql


def query_bigquery(sql_path,
                   sql_params={},
                   location='US'):
    """
    :param sql_path: path to a local .sql file
    :param sql_params: parameters to substitute in the query
    :param location: query location for BigQuery
    :return: pandas dataframe if results=True
    """

    # read the sql, remove the comments:
    sql = read_sql(sql_path, sql_params)

    bigquery_client = bigquery.Client()
    job_config = bigquery.QueryJobConfig()

    query_job = bigquery_client.query(sql,
                                      location=location,
                                      job_config=job_config)

    query_job.result()  # Waits for job to complete.

    return query_job.to_dataframe()


def exec_bigquery(sql_path,
                  dataset,
                  sql_params={},
                  table=None,
                  write_disposition='WRITE_EMPTY',
                  location='US'):

    """
    :param sql_path: path to a local .sql file
    :param sql_params: parameters to substitute in the query
    :param dataset:
    :param table: Name (Optional)
    :param write_disposition: examples:
                'WRITE_TRUNCATE' overwrites anything existing
                'WRITE_APPEND' appends to anything existing
                'WRITE_EMPTY' exits with error if table already exists
    :param location: query location for BigQuery
    :return: pandas dataframe if results=True
    """

    # read the sql, remove the comments:
    sql = read_sql(sql_path, sql_params)

    bigquery_client = bigquery.Client()
    job_config = bigquery.QueryJobConfig()
    job_config.write_disposition = write_disposition

    if table is None:
        table = next(tempfile._get_candidate_names())
    table_ref = bigquery_client.dataset(dataset).table(table)
    job_config.destination = table_ref

    query_job = bigquery_client.query(sql,
                                      location=location,
                                      job_config=job_config)

    query_job.result()  # Waits for job to complete.

    print('Table written to {}'.format(':'.join([dataset, table])))


def export_bigquery(dataset,
                    table,
                    gcs_path,
                    gcs_bucket=os.getenv('GCS_BUCKET'),
                    location='US'):
    """
    Exports data from a BigQuery table to GCS
    :param dataset:
    :param table:
    :param gcs_path:
    :param gcs_bucket:
    :param location:
    :return:
    """

    uri = "gs://{}/{}".format(gcs_bucket, gcs_path)

    bigquery_client = bigquery.Client()
    dataset_ref = bigquery_client.dataset(dataset)
    table_ref = dataset_ref.table(table)

    extract_job = bigquery_client.extract_table(table_ref,
                                                uri,
                                                location=location)

    extract_job.result()  # Waits for job to complete.

    print('Exported {}.{} to {}'.format(dataset, table, uri))


def import_bigquery(dataset,
                    table,
                    gcs_path,
                    gcs_bucket=os.getenv('GCS_BUCKET'),
                    location='US',
                    schema=[],
                    header=True,
                    write_disposition='WRITE_APPEND'):
    """
    Imports data from GCS to a BigQuery table, auto-detecting the schema
    :param dataset: BigQuery dataset to write to
    :param table: BigQuery table to write to
    :param gcs_path: where the GCS CSV resides
    :param gcs_bucket: the GCS bucket
    :param location: data location e.g. 'US'
    :param schema: list of tuples (colname, dtype)
    :param header: does the CSV contain a header?
    :param write_disposition: examples:
                'WRITE_TRUNCATE' overwrites anything existing
                'WRITE_APPEND' appends to anything existing
                'WRITE_EMPTY' exits with error if table already exists
    :return:
    """

    uri = "gs://{}/{}".format(gcs_bucket, gcs_path)

    bigquery_client = bigquery.Client()

    job_config = bigquery.LoadJobConfig()
    if len(schema) == 0:
        job_config.autodetect = True
    else:
        job_config.schema = [bigquery.SchemaField(colname, dtype) for colname, dtype in schema]
        if header:
            job_config.skip_leading_rows = 1

    job_config.source_format = bigquery.SourceFormat.CSV
    job_config.write_disposition = write_disposition

    dataset_ref = bigquery_client.dataset(dataset)
    table_ref = dataset_ref.table(table)

    load_job = bigquery_client.load_table_from_uri(uri,
                                                   table_ref,
                                                   job_config=job_config,
                                                   location=location)

    load_job.result()  # Waits for job to complete.

    print('Imported {} to {}.{}'.format(uri, dataset, table))


def upload_gcs(file_path,
               gcs_path,
               gcs_bucket=os.getenv('GCS_BUCKET')):
    """
    :param file_path: path to file on the local machine
    :param gcs_path: path to write to, within gcs bucket
    :param gcs_bucket: the gcs bucket
    :return:
    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(gcs_bucket)
    blob = bucket.blob(gcs_path)
    blob.upload_from_filename(file_path)


def download_gcs(gcs_path,
                 file_path,
                 gcs_bucket=os.getenv('GCS_BUCKET')):
    """
    :param gcs_path: path to file within the gcs bucket
    :param file_path: path to write to, on the local machine
    :param gcs_bucket: the gcs bucket
    :return:
    """
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(gcs_bucket)
    blob = bucket.blob(gcs_path)
    blob.download_to_filename(file_path)


def df_to_gcs(gcs_path,
              df,
              gcs_bucket=os.getenv('GCS_BUCKET'),
              index=False):
    """
    Writes a Dataframe to Google Cloud Storage directly
    :param gcs_path:
    :param df:
    :param gcs_bucket:
    :return:
    """

    client = storage.Client()
    bucket = client.get_bucket(gcs_bucket)
    bucket.blob(gcs_path).upload_from_string(df.to_csv(index=index), 'text/csv')





import pandas as pd
import numpy as np
import pickle as pkl


class Transformer(object):
    """
    Transforms raw data into modelling data using both declared and trained configurations.
    """
    def __init__(self,
                 min_samples=1000,
                 max_binaries=10,
                 catch_all='other'):

        self.fit_params = dict()

        self.config_params = dict()
        self.config_params['catch_all'] = catch_all
        self.config_params['min_samples'] = min_samples
        self.config_params['max_binaries'] = max_binaries

    def fit(self, df, target=None, binarize=[]):
        """
        :param df: Pandas Dataframe to train on
        :param target: target variable column name
        :param binarize: columns to binarize
        :return: Transformer
        """
        self.fit_params['target_col'] = target
        self.fit_params['columns_in'] = df.columns.tolist()
        self.fit_params['binarize'] = binarize

        binaries = {}
        for col in self.fit_params['binarize']:
            group = df.groupby([col]).agg({col: ['count']})
            group = group[group.iloc[:, 0] >= self.config_params['min_samples']]
            group = group.sort_values(by=group.columns[0], ascending=False)
            binaries[col] = group.iloc[:self.config_params['max_binaries']].index.tolist()
        self.fit_params['binaries'] = binaries

        new_cols = []
        existing_cols = df.columns.tolist()
        for column in self.fit_params['binarize']:
            categories = sorted(self.fit_params['binaries'][column])
            categories.append(self.config_params['catch_all'])
            strcat = [str(int(cat)) if isinstance(cat,(int,np.float)) else cat for cat in categories]
            new_cols.append([column + '_' + cat for cat in strcat])
            existing_cols.remove(column)

        self.fit_params['columns_out'] = existing_cols.append(new_cols)

        return self

    def transform(self, df):
        """
        :param df: Pandas Dataframe to transform
        :return: Transformed Dataframe
        """
        df = df.copy(deep=True)
        for column in self.fit_params['binarize']:
            categories = sorted(self.fit_params['binaries'][column])
            catch_all = self.config_params['catch_all']

            # introduce binary columns, preserving the same order in fit().
            for cat in categories:
                if isinstance(cat, (int, np.float)):
                    catstr = str(int(cat))
                else:
                    catstr = cat
                df['_'.join([column, catstr])] = 0

            df['_'.join([column, catch_all])] = 1

            # update binary columns iteratively
            for cat in categories:
                if isinstance(cat, (int, np.float)):
                    catstr = str(int(cat))
                else:
                    catstr = cat
                df.loc[df[column] == cat, '_'.join([column, catstr])] = 1
                df.loc[df[column] == cat, '_'.join([column, catch_all])] = 0

            # drop categorical column:
            df.drop(column, axis=1, inplace=True)

        return df

    def save(self, pkl_path):
        """
        :param pkl_path: path to save transformer (as pickled dict)
        :return:
        """
        saved_params = dict()
        saved_params['fit_params'] = self.fit_params
        saved_params['config_params'] = self.config_params

        with open(pkl_path, 'wb') as f:
            pkl.dump(saved_params, f)

    @classmethod
    def load(cls, pkl_path):
        """
        :param pkl_path: path to load from (a pickled dict)
        :return: loaded Transformer
        """

        with open(pkl_path, 'rb') as f:
            loaded_params = pkl.load(f)

        temp_cls = Transformer()
        temp_cls.fit_params = loaded_params['fit_params']
        temp_cls.config_params = loaded_params['config_params']

        return temp_cls





import os


def save_requirements(file_path='requirements.txt'):
    cmd = 'pip freeze > {}/{}'.format(os.getcwd(), file_path)
    os.system(cmd)


def install_requirements(file_path='requirements.txt'):
    cmd = 'pip install -r {}/{}'.format(os.getcwd(), file_path)
    os.system(cmd)




from sklearn import metrics
from sklearn.utils import shuffle
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import itertools


def j_score(observed,softprob):
    fpr, tpr, thresholds = metrics.roc_curve(observed,softprob)
    j_scores = tpr - fpr
    return sorted(zip(j_scores,thresholds))[-1][1]


def pr_curve(observed, softprob, title='Precision-Recall Curve', max_precision=1, axes=[0]):
    precision, recall, thresholds = metrics.precision_recall_curve(observed, softprob)
    auprc = metrics.average_precision_score(observed, softprob)
    prev = observed.sum() / observed.count()

    axes.plot([0, 1], [prev, prev], 'r--')
    axes.step(recall, precision, color='b', alpha=0.2, where='post')
    axes.fill_between(recall, precision, color='b', alpha=0.2, step='post')
    axes.set_xlim([0, 1])
    axes.set_ylim([0, max_precision])
    axes.set_xlabel('Recall')
    axes.set_ylabel('Precision')
    axes.set_title(title)
    axes.text(0.01, max_precision, 'AUPRC = %0.2f' % auprc,
              verticalalignment='top',
              horizontalalignment='left',
              fontsize=12)


def roc_curve(observed, softprob, title='Precision-Recall Curve', axes=[0]):
    fpr, tpr, thresholds = metrics.roc_curve(observed, softprob)
    auc = metrics.auc(fpr, tpr)

    axes.plot([0, 1], [0, 1], 'r--')
    axes.plot(fpr, tpr, 'b', label='AUC = %0.2f' % auc)
    axes.set_xlim([0, 1])
    axes.set_ylim([0, 1])
    axes.set_xlabel('False Positive Rate')
    axes.set_ylabel('True Positive Rate')
    axes.set_title(title)
    axes.text(0.01, 1, 'AUC = %0.2f' % auc,
              verticalalignment='top',
              horizontalalignment='left',
              fontsize=12)


def conf_matrix(observed, predicted, axes,
                title=None, class_names=None, normalize=False):

    if class_names is None:
        class_names = list({val for val in observed + predicted})

    if title is None:
        if normalize:
            title = 'Confusion Matrix %ages'
        else:
            title = 'Confusion Matrix'

    observed = [class_names[i] for i in observed]
    predicted = [class_names[i] for i in predicted]
    cm = metrics.confusion_matrix(observed, predicted, class_names)

    if normalize:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

    plt.sca(axes)
    plt.title(title)
    plt.imshow(cm, interpolation='nearest', cmap=plt.get_cmap('Blues'), aspect='auto')
    plt.colorbar(fraction=0.046, pad=0.04, aspect='auto')

    tick_marks = np.arange(len(class_names))
    plt.xticks(tick_marks, class_names)
    plt.yticks(tick_marks, class_names)
    plt.ylabel('True Class')
    plt.xlabel('Predicted Class')

    thresh = cm.max() / 1.5 if normalize else cm.max() / 2
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if normalize:
            plt.text(j, 0.25+i*0.5, "{:0.4f}".format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")
        else:
            plt.text(j, 0.25+i*0.5, "{:,}".format(cm[i, j]),
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")


def regr_mae(df,
             prediction_col,
             observation_col):

    df_copy = df.copy(deep=True)
    df_copy[observation_col] = df_copy[observation_col].fillna(0)
    df_copy[prediction_col] = df_copy[prediction_col].fillna(0)

    df_copy['abs_err'] = abs(df_copy[prediction_col]-df_copy[observation_col])
    return df_copy['abs_err'].sum() / df_copy['abs_err'].count()


def regr_rmse(df,
              prediction_col,
              observation_col):

    df_copy = df.copy(deep=True)
    df_copy[observation_col] = df_copy[observation_col].fillna(0)
    df_copy[prediction_col] = df_copy[prediction_col].fillna(0)

    df_copy['sq_err'] = (df_copy[prediction_col] - df_copy[observation_col])**2
    return np.sqrt(df_copy['sq_err'].sum() / df_copy['sq_err'].count())


def regr_error_within_samples(sample_size,
                              dataset,
                              prediction_col,
                              observation_col,
                              versions_col):
    samples_list = {}
    counter = 0

    for version, chunk in dataset[[versions_col,
                                   prediction_col,
                                   observation_col]].groupby(versions_col):

        chunk.fillna(0, inplace=True)
        chunk = shuffle(chunk)

        for i in range(int(np.floor(chunk.shape[0] / sample_size))):
            sample = chunk[i * sample_size:(i + 1) * sample_size]

            samples_list[counter] = [version,
                                     sample_size,
                                     i,
                                     chunk[observation_col].mean(),
                                     chunk[prediction_col].mean(),
                                     regr_mae(sample, prediction_col, observation_col),
                                     regr_rmse(sample, prediction_col, observation_col)]
            counter = counter + 1

    output = pd.DataFrame.from_dict(samples_list, orient='index')
    output.columns = ['version', 'sample_size', 'sample', 'avg_observed', 'avg_predicted', 'mae', 'rmse']

    grouped = output.groupby(['version', 'avg_observed', 'avg_predicted']).agg({'sample': ['count'],
                                                                                'mae': ['mean', 'max'],
                                                                                'rmse': ['mean', 'max']})

    grouped.columns = ['_'.join(col).strip() for col in grouped.columns.values]
    grouped.reset_index(inplace=True)

    return grouped


def regr_error_by_sample(sample_size,
                         dataset,
                         prediction_col,
                         observation_col,
                         versions_col):
    samples_list = {}
    counter = 0

    for version, chunk in dataset[[versions_col,
                                   prediction_col,
                                   observation_col]].groupby(versions_col):

        chunk.fillna(0, inplace=True)
        chunk = shuffle(chunk)
        n_samples = int(np.floor(chunk.shape[0] / sample_size))

        for i in range(int(np.floor(chunk.shape[0] / sample_size))):
            sample = chunk[i * sample_size:(i + 1) * sample_size]

            samples_list[counter] = [version,
                                     sample_size,
                                     n_samples,
                                     i,
                                     sample[observation_col].sum(),
                                     sample[prediction_col].sum(),
                                     abs(sample[observation_col].sum() - sample[prediction_col].sum())]
            counter = counter + 1

    output = pd.DataFrame.from_dict(samples_list, orient='index')

    output.columns = ['version', 'sample_size', 'num_samples', 'sample_idx', 'sample', 'prediction', 'error']

    grouped = output.groupby(['version', 'num_samples', 'sample_size']).agg({'sample': ['mean'],
                                                                             'error': ['mean', 'max']})

    grouped.columns = ['_'.join(col).strip() for col in grouped.columns.values]
    grouped.reset_index(inplace=True)

    grouped['percent_error_mean'] = grouped['error_mean'] / grouped['sample_mean']
    grouped['percent_error_max'] = grouped['error_max'] / grouped['sample_mean']

    return grouped

def mae_by_date(df,
                versions_col,
                date_col,
                prediction_col,
                observation_col):

    df_copy = df.copy(deep=True)
    df_copy[observation_col] = df_copy[observation_col].fillna(0)
    df_copy[prediction_col] = df_copy[prediction_col].fillna(0)

    df_copy['abs_err'] = abs(df_copy[observation_col] - df_copy[prediction_col])

    accuracy_by_date = df_copy[[versions_col, date_col, 'abs_err']].pivot_table(index=[date_col],
                                                                                values=['abs_err'],
                                                                                columns=[versions_col],
                                                                                aggfunc=np.mean)

    accuracy_by_date.columns = [col[1] for col in accuracy_by_date.columns.tolist()]
    accuracy_by_date.reset_index(inplace=True)

    return accuracy_by_date
